
import pandas as pd
from optparse import OptionParser
from sklearn.preprocessing import MinMaxScaler
from sklearn.model_selection import StratifiedKFold
from torch import nn
import pytorch_lightning as pl
from pytorch_lightning.callbacks.early_stopping import EarlyStopping
from pytorch_lightning.loggers import TensorBoardLogger
from csv import DictWriter
from MyDataTransforms import DataModule
from ANN_Classification_Modules import GEClassifier_2layer

########################################################################
########################################################################

parser = OptionParser()

parser.add_option("-f", dest = "fold", default = "", help = "which fold ist test set")
parser.add_option("-g", dest = "genes", default = "", help = "number of genes to include")
options, args = parser.parse_args()

f = int(options.fold)
g = int(options.genes)

tb_logger_folder = '{}_classification_logs'.format(method)
output_filename = 'methods_modelperformances_on_test_subsets.csv'


#######################################################################
############################# IMPORT DATA #############################
#######################################################################

# Import list of p-val sorted genes from respective cross validation fold
# only first 'g' genes should be included in the model
genes = pd.read_csv('your_file_directory/sortedgenes_fold{}.csv'.format(f), index_col = 0).index.tolist()[:g]

# Import augmened data set including augmented training samples for both classes
# as well as test samples
# Besides expression values the data set should include columns for 'subset' identification
# ('train' or 'test') and 'label' identification (0 or 1)
DF = pd.read_csv('your_file_directory/mwoAugmented_train_test_fold{}.csv'.format(f), index_col='id')
DF = DF[['subset','label']+genes]

# split data
train = DF[(DF.subset == 'train')]
test = DF[(DF.subset == 'test')]


######################################################################
############################# SCALE DATA #############################
######################################################################

scaler = MinMaxScaler(feature_range=(-1, 1))
scaler.fit(train[genes])

train_sc = scaler.transform(train[genes])
train_sc = pd.DataFrame(train_sc, index = train.index, columns = genes)
train_sc.insert(0, 'label', train.label)

test_sc = scaler.transform(test[genes])
test_sc = pd.DataFrame(test_sc, index = test.index, columns = genes)
test_sc.insert(0, 'label', test.label)


###############################################################################
############################# SPLIT TRAINING DATA #############################
###############################################################################

# create training-internal validation set (10%)
train_sc = train_sc.sample(frac=1)
kfold = StratifiedKFold(n_splits=10, random_state = None, shuffle=False)
train_ids, val_ids = next(iter(kfold.split(train_sc, train_sc.label)))


###################################################################################
############################# SETTINGS FOR CLASSIFIER #############################
###################################################################################

max_epochs = 200
min_epochs = 50

criterion = nn.BCEWithLogitsLoss()

classifier = GEClassifier_2layer(g, criterion)  

early_stop_callback = EarlyStopping(monitor='val_loss', patience=20, verbose=False, mode='min')

tblogger = TensorBoardLogger(tb_logger_folder, name='{}_{}topgenes_fold{}'.format(classifier.__class__.__name__, g, f))

trainer = pl.Trainer(max_epochs = max_epochs, min_epochs = min_epochs, gpus=1, progress_bar_refresh_rate = 10, 
			callbacks = [early_stop_callback], logger = tblogger)

dm = DataModule(train = train_sc.iloc[train_ids,:], 
		val = train_sc.iloc[val_ids,:], 
		test = test_sc, 
		genes = genes)


############################################################################
############################# TRAIN CLASSIFIER #############################
############################################################################

trainer.fit(classifier, dm)


###########################################################################
############################# TEST CLASSIFIER #############################
###########################################################################

# test on test data 
trainer.test(datamodule=dm)

# save measures in csv file
metrics = trainer.test(datamodule=dm)[0]
info = {'numTopGenes': g, 'cvFold': f}
z = {**info, **metrics}
colnames = ['numTopgenes', 'cvFold', 'acc', 'bs', 'sens', 'spec', 'tp', 'tn', 'fp', 'fn']

with open (output_filename, 'a') as eval_object:
	dictwriter_object = DictWriter(eval_object, fieldnames = colnames)
	dictwriter_object.writerow(z)
	eval_object.close()




	
