import torch
import pytorch_lightning as pl
from torch import nn
from sklearn.metrics import confusion_matrix
import numpy as np
import torchmetrics


stat_scores = torchmetrics.StatScores().to('cuda')
accuracy = torchmetrics.Accuracy().to('cuda')
precision = torchmetrics.Precision().to('cuda')
recall = torchmetrics.Recall().to('cuda')



def reset_weights(m):
	for layer in m.children():
		if hasattr(layer, 'reset_parameters'):
			print(f'Reset trainable parameters of layer = {layer}')
			layer.reset_parameters()
	

class GEClassifier_1layer(pl.LightningModule):
	def __init__(self, g, criterion):
		super().__init__()
		self.g = g
		self.layers = nn.Sequential(
			nn.Linear(self.g, 75),
			#nn.LeakyReLU(0.2),
			nn.Tanh(),
			nn.Dropout(0.2),
			nn.Linear(75, 1),
		)
		self.criterion = criterion
	def forward(self, x):
		return self.layers(x)
	def training_step(self, batch, batch_idx):
		x,y = batch
		x = x.view(x.size(), -1)
		y_flt = y.float()
		logit = self.layers(x).flatten()
		loss = self.criterion(logit, y_flt)
		y_hat = torch.round(torch.sigmoid(logit))
		acc = (y_hat == y).sum().float()/y.shape[0]
		self.log('train_loss', loss, on_step=True, on_epoch=True, prog_bar=True, logger=True)
		self.log('train_acc', acc, on_step=True, on_epoch=True, prog_bar=True, logger=True)
		return loss
	def validation_step(self, batch, batch_idx):
		x, y = batch
		x = x.view(x.size(), -1)
		y_flt = y.float()
		logit = self.layers(x).flatten()
		loss = self.criterion(logit, y_flt)
		y_hat = torch.round(torch.sigmoid(logit))
		acc = (y_hat == y).sum().float()/y.shape[0]
		self.log('val_loss', loss, prog_bar=True)
		self.log('val_acc', acc, prog_bar=True)
		return loss
	def configure_optimizers(self):
		optimizer = torch.optim.RMSprop(self.parameters(), lr=0.01)
		scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode = 'min', factor=0.2)
		return {'optimizer': optimizer,'scheduler':scheduler,'monitor':"val_loss",}
	def test_step(self, batch, batch_idx):
		x, y = batch
		x = x.view(x.size(0), -1)
		y_flt = y.float()
		logit = self.layers(x).flatten()
		loss = self.criterion(logit, y_flt)
		bs = (torch.subtract(torch.sigmoid(logit), y)**2).sum()/len(y)
		y_hat = torch.round(torch.sigmoid(logit))
		acc = accuracy(y_hat, y)
		#prec = precision(y_hat, y)
		#rec = recall(y_hat, y)
		tp = stat_scores(y_hat, y)[0].item() 
		fp = stat_scores(y_hat, y)[1].item()
		tn = stat_scores(y_hat, y)[2].item()
		fn = stat_scores(y_hat, y)[3].item()
		sens = tp / (tp + fn)
		spec = tn / (tn + fp) 
		self.log('acc', acc)
		self.log('bs', bs)
		self.log('sens', sens)
		self.log('spec', spec)
		self.log('tp', tp)
		self.log('tn', tn)
		self.log('fp', fp)
		self.log('fn', fn)		
		return loss
		
		
		
class GEClassifier_2layer(pl.LightningModule):
	def __init__(self, g, criterion):
		super().__init__()
		self.g = g
		self.layers = nn.Sequential(
			nn.Linear(self.g, 150),
			nn.LeakyReLU(0.2),
			nn.Dropout(0.2),
			nn.Linear(150, 50),
			nn.LeakyReLU(0.2),
			nn.Dropout(0.2),
			nn.Linear(50, 1),
		)
		self.criterion = criterion
	def forward(self, x):
		return self.layers(x)
	def training_step(self, batch, batch_idx):
		x,y = batch
		x = x.view(x.size(), -1)
		y_flt = y.float()
		logit = self.layers(x).flatten()
		loss = self.criterion(logit, y_flt)
		y_hat = torch.round(torch.sigmoid(logit))
		acc = (y_hat == y).sum().float()/y.shape[0]
		self.log('train_loss', loss, on_step=True, on_epoch=True, prog_bar=True, logger=True)
		self.log('train_acc', acc, on_step=True, on_epoch=True, prog_bar=True, logger=True)
		return loss
	def validation_step(self, batch, batch_idx):
		x, y = batch
		x = x.view(x.size(), -1)
		y_flt = y.float()
		logit = self.layers(x).flatten()
		loss = self.criterion(logit, y_flt)
		y_hat = torch.round(torch.sigmoid(logit))
		acc = (y_hat == y).sum().float()/y.shape[0]
		self.log('val_loss', loss, prog_bar=True)
		self.log('val_acc', acc, prog_bar=True)
		return loss
	def configure_optimizers(self):
		optimizer = torch.optim.SGD(self.parameters(), lr = 1e-3)
		scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode = 'min', factor=0.2)
		return {'optimizer': optimizer,'scheduler':scheduler,'monitor':"val_loss",}
	def test_step(self, batch, batch_idx):
		x, y = batch
		x = x.view(x.size(0), -1)
		y_flt = y.float()
		logit = self.layers(x).flatten()
		loss = self.criterion(logit, y_flt)
		bs = (torch.subtract(torch.sigmoid(logit), y)**2).sum()/len(y)
		y_hat = torch.round(torch.sigmoid(logit))
		acc = accuracy(y_hat, y)
		tp = stat_scores(y_hat, y)[0].item() 
		fp = stat_scores(y_hat, y)[1].item()
		tn = stat_scores(y_hat, y)[2].item()
		fn = stat_scores(y_hat, y)[3].item()
		sens = tp / (tp + fn)
		spec = tn / (tn + fp) 
		self.log('acc', acc)
		self.log('bs', bs)
		self.log('sens', sens)
		self.log('spec', spec)
		self.log('tp', tp)
		self.log('tn', tn)
		self.log('fp', fp)
		self.log('fn', fn)		
		return loss
