import pandas as pd
import random
from sklearn.preprocessing import MinMaxScaler
import torch
import torch.optim as optim
from torch import nn
from torch.utils.data import DataLoader
from torch.autograd.variable import Variable
from MyDataTransforms import GeneExpressionDataset
from optparse import OptionParser

parser = OptionParser()
parser.add_option("-f", dest = "fold", default = "", help = "Welches Fold wird benutzt")
options, args = parser.parse_args()

label = 1               # class for which the data is to be augmented
method = 'GAN'          # augmentation method
f = int(options.fold)   # fold of cross validation
augFactor = 5           # augmentation factor 


ngpu = 1        		# number of gpus
workers = 16     		# number of data loading workers
z_dim = 100     		# size of the latent vector z 
batchSize = 2   		# batch size
niter = 2000     		# number of epochs to train for
learning_rate = 1e-4    # learning rate
scaler = MinMaxScaler(feature_range=(-1, 1))  


################################################################
###################### LOAD ORIGINAL DATA ######################
################################################################

# Load file including training and test data for the specific k-fold split 
# Besides gene expression values data set has to contain information about 
# the subset type ('train' or 'test') as well as the labels of the samples (0 or 1)
DF = pd.read_csv('your_file_directory/train_test_split_fold{}.csv'.format(f), index_col=0)

# only include training data from specified class (label)
gandata = DF[(DF.subset == 'train') & (DF.label == label)]

# select expression values
genes = gandata.select_dtypes('float').columns
X_dim = len(genes)

print('Generate new data for label {} ; train for {} epochs'.format(gandata.label.unique().item(), niter))

###################################################
###################### MODEL ######################
###################################################

class Generator(nn.Module):
    def __init__(self):
        super(Generator, self).__init__()
        self.main = nn.Sequential(
            nn.Linear(z_dim, 250, bias=True),
            nn.LeakyReLU(0.2,inplace=True),
            nn.Linear(250, 500, bias=True),
            nn.LeakyReLU(0.2,inplace=True),
            nn.Linear(500, 1000, bias=True),
            nn.LeakyReLU(0.2,inplace=True),
            nn.Linear(1000, X_dim, bias=True),
            nn.Tanh()
        )
    def forward(self, input):
        output = self.main(input)
        return output


class Discriminator(nn.Module):
    def __init__(self):
        super(Discriminator, self).__init__()
        self.main = nn.Sequential(
            nn.Linear(X_dim, 1000),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(1000, 500),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(500, 250),
            nn.LeakyReLU(0.2, inplace=True),
            nn.Linear(250, 1),
        )
    def forward(self, input):
        output = self.main(input)
        return output.squeeze(1)


def weights_init(m):
    classname = m.__class__.__name__
    if isinstance(m, nn.Linear):
        nn.init.normal_(m.weight, 0.0, 0.02)

generator = Generator().cuda()
generator.apply(weights_init)
       
discriminator = Discriminator().cuda()
discriminator.apply(weights_init)
   	
optimizerD = optim.Adam(discriminator.parameters(), lr=0.0001)
optimizerG = optim.Adam(generator.parameters(), lr=0.0001)

loss = nn.BCEWithLogitsLoss()
      


########################################################################
###################### TRAINING / DATA GENERATION ######################
########################################################################

# Scale values to range (-1, 1)
scaler.fit(gandata[genes])
gandata_sc = scaler.transform(gandata[genes])

# Prepare data for later saving
gandata_sc = pd.DataFrame(gandata_sc, index = gandata.index, columns = genes)
gandata_sc.insert(0, 'label', gandata.label)
gandata_sc.insert(0, 'id', gandata.id)
gandata_sc.insert(0, 'subset', gandata.subset)

# Transfer the dataset into a specific structure that can be used by the pytorch dataloader for training
dataset = GeneExpressionDataset(gandata_sc, genes, True)

# Create pytorch DataLoader object
dataloader = DataLoader(dataset, batch_size = 2, shuffle = True, num_workers = 16)

# Specify number of samples that should be simulated
nnew = len(gandata)*augFactor

# create matrix containing nnew rows times z_dim columns
fixed_noise = torch.randn(nnew, z_dim).cuda()


# Training of generator and discriminator networks

generator.train()
discriminator.train()

for epoch in range(niter):
	for i, data in enumerate(dataloader, 0):
		real = data[0].cuda()
		N = len(real)
		noise = torch.randn(N, z_dim).cuda()
		fake = generator(noise) 
		discriminator.zero_grad()
		d_output_real = discriminator(real)
		d_error_real = loss(d_output_real, Variable(torch.ones(N)*random.uniform(0.9, 1.0)).cuda())
		d_error_real.backward()
		d_output_fake = discriminator(fake)
		d_error_fake = loss(d_output_fake, Variable(torch.ones(N)*random.uniform(0.0, 0.1)).cuda())
		d_error_fake.backward()
		optimizerD.step()
		d_error = d_error_real + d_error_fake
		generator.zero_grad()
		fake = generator(noise)
		d_output_fake = discriminator(fake)
		g_error = loss(d_output_fake, Variable(torch.ones(N)).cuda())
		g_error.backward()
		optimizerG.step()
		print('[%d/%d][%d/%d] Loss_D: %.4f Loss_G: %.4f' % (epoch, niter, i, len(dataloader), d_error, g_error))


# generate artificial samples
with torch.no_grad():
	fakes = generator(fixed_noise).detach().cpu()

# re-transform scaled values
gen_np = fakes.numpy()
gen_np = scaler.inverse_transform(gen_np)

# save generated samples in dataframe
gen_df = pd.DataFrame(gen_np)
gen_df.columns = genes

# insert info columns for later analyses
idxnames = ['a{}_label{}'.format(i+1, label) for i in range(nnew)]
gen_df.insert(0, 'label', [label]*nnew)
gen_df.insert(0, 'id', idxnames)
gen_df.insert(0, 'subset', ['train']*nnew)

# append generated samples to original samples and save as csv file
columns = ['subset','id','label']+genes.tolist()
aug_df = DF[(DF.label == label)][columns].append(gen_df[columns])
aug_df.to_csv('{}_{}iterations_SubsetLabel{}_5times_augmented_train_test_fold{}.csv'.format(method, niter, label, f), index=False)



